#include <fstream.h>
#include <limits.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>

#include "FC.h"
#include "MyTypes.h"
#include "AudioDriver.h"

#if defined(HAVE_STRNCASECMP)
  #undef strnicmp
  #define strnicmp strncasecmp
#endif

const udword extraFileBufLen = 8+1;  // see FC.cpp

bool exitFlag;

void (*oldSigHupHandler)(int);
void (*oldSigIntHandler)(int);
void (*oldSigQuitHandler)(int);
void (*oldSigTermHandler)(int);
void mysighandler(int signum)
{
    switch (signum)
    {
     case SIGHUP:
        {
            exitFlag = true;
            break;
        }
     case SIGINT:
        {
            exitFlag = true;
            break;
        }
     case SIGQUIT:
        {
            exitFlag = true;
            break;
        }
     case SIGTERM:
        {
            exitFlag = true;
            break;
        }
     default:
        break;
    }
}

int main(int argc, char *argv[])
{
	// File argument number.
	int inFileArgN = 0;
    int startStep = 0;
    int endStep = 0;
    
    // Parse command line arguments.
    for ( int a = 1; a < argc; a++)
    {
        if ( argv[a][0] == '-')
        {
            if ( strnicmp(&argv[a][0],"--start=",8) == 0)
            {
                startStep = atoi(argv[a]+8);
            }
            else if ( strnicmp(&argv[a][0],"--end=",6) == 0)
            {
                endStep = atoi(argv[a]+6);
            }
        }
		else
		{
            // Set filename argument number.
            if (inFileArgN == 0)
            {
                inFileArgN = a;
            }
            else
            {
                cerr << argv[0] << ": Missings argument(s)." << endl;
            }
        }
    }

    if (inFileArgN == 0)
    {
        cerr << argv[0] << ": Missing file name argument." << endl;
        exit(-1);
    }

    ubyte* buffer = 0;
    streampos fileLen = 0;
  
    // open binary input file stream at end of file
#if defined(HAVE_IOS_BIN)
    ifstream myIn( argv[inFileArgN], ios::in | ios::bin | ios::ate | ios::nocreate );
#else
    ifstream myIn( argv[inFileArgN], ios::in | ios::binary | ios::ate | ios::nocreate );
#endif
	// As a replacement for !is_open(), bad() and the NOT-operator don't seem
	// to work on all systems.
#if defined(DONT_HAVE_IS_OPEN)
    if ( !myIn )
#else
	if ( !myIn.is_open() )
#endif
    if ( !myIn )
    {
        cerr << argv[0] << ": ERROR: Cannot open file: " << argv[inFileArgN] << endl;
        return(-1);
    }
#if defined(HAVE_SEEKG_OFFSET)
    fileLen = (myIn.seekg(0,ios::end)).offset();
#else
    myIn.seekg(0,ios::end);
    fileLen = (udword)myIn.tellg();
#endif
	fileLen = myIn.tellg();
	buffer = new ubyte[fileLen+extraFileBufLen];
	
	myIn.seekg(0,ios::beg);
	cout << "Loading" << flush;
	streampos restFileLen = fileLen;
	while ( restFileLen > INT_MAX )
    {
        myIn.read( buffer + (fileLen - restFileLen), INT_MAX );
        restFileLen -= INT_MAX;
        cout << "." << flush;
	}
	if ( restFileLen > 0 )
    {
        myIn.read( buffer + (fileLen - restFileLen), restFileLen );
        cout << "." << flush;
	}
	cout << endl << flush;
	
	if ( myIn.bad() )
    {
        cerr << argv[0] << ": ERROR: Cannot read file: " << argv[inFileArgN] << endl;
        return(-1);
    }
	myIn.close();
  
    extern udword pcmFreq;
    pcmFreq = 44100;

    ubyte* sampleBuffer = new ubyte[pcmFreq];

    AudioConfig myAudioCfg;
    myAudioCfg.frequency = pcmFreq;
    myAudioCfg.precision = AudioConfig::BITS_8;
    myAudioCfg.channels = AudioConfig::MONO;
    myAudioCfg.encoding = AudioConfig::UNSIGNED_PCM;
    
    myAudioCfg.maxFrags = 16;
    myAudioCfg.fragSize = 512; // 256;
        
    exitFlag = false;
    if ((signal(SIGHUP,&mysighandler) == SIG_ERR)
        || (signal(SIGINT,&mysighandler) == SIG_ERR)
        || (signal(SIGQUIT,&mysighandler) == SIG_ERR)
        || (signal(SIGTERM,&mysighandler) == SIG_ERR))
    {
        cerr << argv[0] << ": ERROR: Cannot set signal handlers." << endl;
        exit(-1);
    }

    AudioDriver myAudio;
    if ( !myAudio.open(myAudioCfg) )
        exit(-1);
    else
    {
        if ( myAudioCfg.precision == AudioConfig::BITS_16 )
        {
            cerr << argv[0] << ": ERROR: 16-bit not implemented." << endl;
            myAudio.close();
            exit(-1);
        }
        if (FC_init(buffer,fileLen,startStep,endStep))
        {
            cout << "Module format: " << mixerFormatName << endl;
            
            // Define the ``zero'' sample for signed or unsigned samples.
            ubyte zero8bit = 0x80;
	
            if ( myAudioCfg.precision == AudioConfig::BITS_8 )
            {
                switch( myAudioCfg.encoding )
                {
                 case AudioConfig::SIGNED_PCM:
                    {
                        zero8bit = 0;
                        break;
                    }
                 case AudioConfig::UNSIGNED_PCM:
                 default:
                    {
                        zero8bit = 0x80;
                        break;
                    }
                }
            } // end else myAudioCfg.precision
            
            extern void mixerInit(ubyte);
            mixerInit(zero8bit);
            extern void mixerSetReplayingSpeed();
            mixerSetReplayingSpeed();
            
            cout << "Buffers      : " << myAudio.getConfig().fragSize
                << "\nBuffersize   : " << myAudio.getConfig().blockSize << "\n";
            cout << "Frequency    : " << dec << myAudio.getConfig().frequency << " Hz" << endl;
            while (true)
            {
                extern void mixerFillBuffer( void*, udword );
                mixerFillBuffer(sampleBuffer,myAudio.getConfig().blockSize);
                myAudio.play(sampleBuffer,myAudio.getConfig().blockSize);
                if (exitFlag)
                    break;
            }
        }
        else
        {
            cout << "File format not recognized." << endl;
        }
    }
    
    delete[] sampleBuffer;
        
    if ( buffer != 0 )
    {
        delete[] buffer;
        buffer = 0;
    }

    return(0);
}
